=head Utility Routines for mySQL connections

This is a set of miscellaneous routines for manipulation of the 9-mer database

Written by Phillip J Wyss, Purdue University, Department of Chemistry

Written 10/5/1

Modifications

10/15/1 Use STDERR for prompt rather than default STDOUT

10/11/1 Add Mer2RFmap routine

10/11/1	Move binary search routine (findMerIndex) to this library

10/10/1 Add FivePrimeComplement routine

11/13/1 Add findRFsource routine

11/14/1 Limit lookup in Mer2RFmap by hit count if hit count limit passed

1/17/2  Re-vamp for use with GenBank format via BioPerl rather than XML

1/22/2  Add fullFileName to expand root file name to full pathname

1/23/2  Add computeRawCoordinates to translate TI origin coordinates
        to sequence coordinates

1/23/2  Speed up lookup of 9mer index by via computation of index 
        and cache-only of data

2/26/2	Incorporate compute_mer and compute index with a standard
	transformation of nucleotide to base-4 digit

3/5/2	Add checkEnvironment to avoid redundent code in all utilities

3/6/2   Add findEntity to make sure class name and type are defined

3/10/2  add loadQuery to enable other output beside RF_id,sequence,hits

5/14/2  add getFeatures to extract feature locations from sequence object

5/24/2  add getLocation to get location of single feature, have getFeatures
        call get Location

5/22/2  add pathFind to look for files in path

6/25/3  add possible query for updateHits

11/1/04 cut getFeatures/getLocation from the module (uses obsolete BioPerl module

=cut


package Util9mer;

use strict;
require DBI;
require Date::Manip;

require Bio::Seq;
#obsolete require Bio::Annotation;
require Bio::Location::Simple;
require Bio::Location::Split;

use vars qw( $VERSION @ISA @EXPORT @EXPORT_OK );

use Exporter ();
@ISA = qw(Exporter);
@EXPORT_OK = qw( &Connect2DataSource &FivePrimeComplement &findMerRef &Mer2FRmap $merRefStmt &compute_mer &compute_index &findEntity &loadQuery &setQuery);

use vars qw( $lastError $defaultFormat $base_order $merRefStmt $merRefSQLcmd @_merRefSQLparameters %_ENVERR);

my $saveINT;

$defaultFormat = 'GenBank'; # default format for sequence files

# default sql command for finding hit count for given 9mer

$merRefSQLcmd = '';

# state error messages given if environment variable not found

%_ENVERR = (	'DBI_DSN','Please specify datasource name as DSI_DSN',
		'DBI_USER','Please specify user name as DSI_USER'
	);
#
# data for converting sequence to a number
#

$base_order = 'ATCG';

#
# subroutine for catching the SIGINT signal (not for export)
#
# Input
#
#	The signal name passed is ignored 
#	The routine assumes this was called as the result of a SIGINT
#	which is normally generated by typing a CTRL-C
#
# Result
#
#	The routine makes sure the echo is turned on for the terminal
#       and then exits gracefully
#

sub _vector_c {

	system('stty echo');
	if( $saveINT ) {
		&$saveINT(@_);
	}
	else {
		exit;
	}

}

=head1 subroutine Connect2DataSource - provide a DBI connection to specified database

 Input

	name of data source name without DBI: appended
	user account to make connection
	user password is taken from STDIN

 Result

	a DBI connection object is created and returned
	the routine dies ingratiously if creation of the object fails
	at best it returns a undef-ed value on failure

=cut

sub Connect2DataSource {

#
# get the subroutine parameters
#

	my ($datasource,$user_account) = @_;
	my $dbh; # returned value
#
# turn off console echo for password entry then prompt and input password
# remembering to turn echo back on and removing line-feed from input
#

	my $savedINT = $SIG{INT}; # if another module has dibs on SIGINT
	$SIG{INT} = \&_vector_c;   # execute this first

if( !exists($ENV{'MYSQL_PWD'}) ) {
	system('stty -echo');
	print STDERR "Password: ";
	my $user_password = <STDIN>;
	system('stty echo');

	chomp $user_password; # remove LF

	print STDERR "\n"; # really did want to echo the CRLF
 
	$SIG{INT} = $saveINT;	# restore who had dibs on SIGINT before
	$saveINT = undef;	# this module doesn't have to call it

	$dbh = DBI->connect("dbi:$datasource",$user_account,$user_password);
}
else {
	$dbh = DBI->connect("dbi:$datasource",$user_account);
}
	return $dbh;

}

=head1 subroutine mkSQLdatetime - format datetime for SQL INSERT

Input

	seconds past the current epoch

Result

	a string of the format YYYY-MM-DD HH:MM:SS is returned

=cut

sub mkSQLdatetime {

	return Date::Manip::UnixDate(scalar localtime(shift),'%Y-%m-%d %H:%M:%S');

}

=head1 FivePrimeComplement - Find sequence complement in 5' canonical order

Input

	A string representing the sequence (ATCG)

Results

	A string representing the complement is returned

=cut

sub FivePrimeComplement {

#
# now here is a clever way to do it in Perl
#

	my $seq = shift;
	$seq =~ tr/ATCG/TAGC/; # complement
	return join '', reverse split / */, $seq; # change 3'-> 5' to 5' -> 3'

}

=head1 compute_index - compute_index - compute an array index from mer

Input

	$mer - the string of bases to compute the index for

Results

	The string is evaluated as a base-4 number with each nucleotide
	equated with a digit 0-3.

Returned Value

	The base-4 value computed

=cut

sub compute_index {

	my $mer = shift; # what to compute index from
	my $mer_index = 0; # add each base in base 4

	for( my $i=0; $i < length($mer); $i++) {
		$mer_index = $mer_index*4 + index($base_order,substr($mer,$i,1));
	}

	return $mer_index;

}

=head1 compute_mer - transform an index into a nucleotide sequence

Input

	the index to transform
	the size of the sequence to generate (0 could mean A,AA,AAA,etc)

Results

	The index is represented as a base-4 number with 0,1,2,3 digits
	transformed to A,T,C,G 

Returned Value

	The string of nucleotide bases

=cut

sub compute_mer {

	my $mer_index = shift; # index to decode
	my $mer_size = shift; # how long a mer to generate
	my $mer = ''; # no bases yet

	for(my $i = 0; $i < $mer_size; $i++) {

		$mer = substr($base_order,$mer_index%4,1).$mer;
		$mer_index = int($mer_index/4);

	}

	return $mer;

}

=head1 loadQuery - read a query from a file

Input

	query file name (can be augmented by STORED_PROCEDURES)

Results

	The query file is read to get the sql command of interest

Returned Value

	The query string

=cut

sub loadQuery {


	my $sqlcmd;

	if( $_[0] eq '-' ) { # query input is standard input
		
		$sqlcmd = <STDIN>;
		while(<STDIN>) {
			$sqlcmd .= ' '.$_;
		}
		if( !$sqlcmd ) {
			print STDERR "No query on standard input";
		}
	}
	else {

		my $source_file = findPath($_[0],'STORED_PROCEDURES');

#
# open the file and concatenate all the lines into one big sql statement
#

		open QUERY,"<$source_file";

		$sqlcmd = <QUERY>;
		while(<QUERY>) {
			$sqlcmd .= ' '.$_;
		}

		close QUERY;
		if( !$sqlcmd ) {
			$lastError = "Failed to read query from $source_file";
		}

	}

	return $sqlcmd;

}

=head1 setQuery - set query command and parameters for findMerRef

Input

	query command
	parameter 2
	parameter 3
	.
	.
	.

Results

	The query cmd and parameters are saved
	in package variables for later use.

Return Value

	1 if all successful, 0 if not

=cut

sub setQuery {

	$merRefSQLcmd = shift;
	my $status = 1;
	if( $lastError ) {
		$status = 0;
	}
	@_merRefSQLparameters = @_;

	return $status;

}

=head1 findMerRef subroutine for finding information about oligomer

 Input

	oligomer as string
	reference to an array of array references
	  each of the later references referring to information
          from query record(s)
	database reference from which to fetch non-cached data

 Results

	The index where the sequence was found is returned
	-1 is returned if the query produced no record or error

=cut

sub findMerRef {

	my $seqString = shift; #sequence to look for
	my $seqref = shift;   # reference to the array to look in
	my $dbh = shift; 

	my $currentndx = compute_index(uc $seqString);

#
# check if we have already fetched the mer record(s) in question
# and do so if not
#

	if( ! $$seqref[$currentndx] ) {
		if( !$merRefStmt ) {
			$merRefStmt = $dbh->prepare($merRefSQLcmd);
		}
		if( $merRefStmt ) {
			$merRefStmt->execute($seqString,@_merRefSQLparameters);
			if( $merRefStmt->err ) {
				$lastError = $merRefStmt->errstr;
				$currentndx = -1;
			}
			else {
#
# must touch the attribute so that it persists
#
				my $touch = $merRefStmt->{'NAME'};
#
# get the records and copy to new anonymous array
#

				my $recordref = $merRefStmt->fetchall_arrayref;
				if( ! @$recordref ) { # no rows
					$currentndx = -1;
				}
#
# single array ref for multiple records
#
				if(scalar @$recordref == 1) {
					my $single = $$recordref[0];
					my $copy = [];
					@$copy = @$single; # recordref gets reused
					$$seqref[$currentndx] = $copy;
				}
#
# array ref of array refs for multiple records
#
				if(scalar @$recordref > 1) {
					my $copy = [];
					@$copy = @$recordref;
					$$seqref[$currentndx] = $copy;
				}
			}
		}
		else {
			$currentndx = -1;
		}
	}					

	return $currentndx;

}

=head1 Subroutine findRFsource - get list of reference genes matching a 9mer

Input

	RF id of the 9mer
	the database handle for the reference genes 

Result

	The database experiment table is searched for
	the corresponding location hit records

Returned Value

	a reference to a list of list references
	each reference in the list of references
        reflects the experiment records for the reference genes
	the order of the data is Location,Direction and Gene_labname

=cut

sub findRFsource {

	my $RFid = shift; # the RFid to lookup
	my $dbh = shift; # the database handle to use in generating statement

	my $result = []; # empty reference to list is returned on failure

	my $stmt = $dbh->prepare('SELECT Location,Direction,Sequence_labname FROM RF_code_locate,Sequence_name WHERE Sequence_name.Sequence_id = RF_code_locate.Sequence_id AND RF_id = ?');

	if( $stmt ) {

		$stmt->execute($RFid);
		if( !$stmt->err ) {
			$result = $stmt->fetchall_arrayref;
		}
		else {
			$lastError = $stmt->errstr;
		}

	}

	return $result;

}
		
=head1 fullFileName - convert base file name to full path name

Input

	base file name
	environment variable name to use as base directory
	(optional, default:SEQUENCE_ARCHIVE)
	if value of environment variable looks like a path,
	the first directory is taken

Result

	If the file is already a absolute path name ( starts with . or /)
	no path is prefixed to the file.

	If file is to be prefixed, it is taken from the value of the
	environment variable SEQUENCE_ARCHIVE.

	If SEQUENCE_ARCHIVE is not defined, no prefix given
	
Return value

	full file name to be used

=cut

sub fullFileName {

	my $basename = shift; # base name should always be present

	my $directoryEnvironment = 'SEQUENCE_ARCHIVE';

	if( @_ ) {
		$directoryEnvironment = shift;
	}

# if nothing else

	my $fullname = $basename;

# if not absolute tack on archive directory if directed by env variable

	if( $basename =~ /^[^\.\/]/ ) {
		if( exists $ENV{$directoryEnvironment}) {
			my @directories = split /[:;]/,$ENV{$directoryEnvironment};
			$fullname = $directories[0].'/'.$fullname;
		}
	}

	return $fullname;

}

=head1 computeRawCoordinates - translate TI coordinates to raw sequence ones

Input

	transcription initiation site (TI site)
	beginning coordinte of subsequence
	ending coordinate of subsequence

Results

	Given a transcription initiation site, take coordinates wrt
	to that origin and compute sequence coordinates where position 1
	is first item of sequence

	The TI site is stated wrt the raw sequence coordinates

Returned value

	a two-item list giving beginning,ending coordinate is sequence 
	coordinate system

=cut

sub computeRawCoordinates {

	my $ti = shift; # get origin of transcription coordinates
	my $begin = shift; # beginning of subsequence region
	my $end = shift; # end of subsequence region

	if( !$ti ) { # $ti == 0 means already sequence coordinates
		$ti=1;
	}

# since there is no 0 position in sequence, we need to adjust differently
# for negative and positive coordinates

	if( $begin < 0 ) {
		$begin = $ti+$begin;
	}
	else {
		$begin = $ti+$begin-1;
	}
	if( $end < 0 ) {
		$end = $ti+$end;
	}
	else {
		$end= $ti+$end-1;
	}

	return ($begin,$end);

}

=head1 updateHits - update hit count in RF_code_hits from RF_code_locate

Input

	database handle for database connection for table access

Results

	The hit count in RF_code_hits is updated to reflect the count from
	RF_code_locate

Returned Value

	zero for failure, non-zero for success

=cut

sub updateHits {

	my $dbh = shift;
	my $updateQuery = 'SELECT COUNT(RF_id),RF_id AS Hits FROM RF_code_locate GROUP BY RF_id';
	my @parameters = ();
	if( @_ ) {
		$updateQuery = shift;
		@parameters = @_;
	}
	my $status = 1; # change to zero on error
	my $result = 0; # change to one on final updates

#
# try getting all RF_id's along with their hit count calculated from
# the number of location records available for each
#

	my $stmt = $dbh->prepare($updateQuery);
	if($dbh->err) {
		$status = 0;
		$lastError = $dbh->errstr;
	}
	else {
		$stmt->execute(@parameters);
		if( $stmt->err ) {
			$lastError = $stmt->errstr;
			$status = 0;
		}
	}
			
#
# zero out all current counts
#

	if( $status ) {
		my $update = $dbh->prepare('UPDATE RF_code_hits SET Num_of_hits = 0');
		if( $dbh->err) {
			$lastError = $dbh->errstr;
			$status = 0;
			$stmt->finish();
		}
		else {
			$update->execute();
			if( $update->err ) {
				$lastError = $update->errstr;
				$status = 0;
				$stmt->finish();
			}
		}
		$update->finish();
	}

#
# now copy the count for each id to the RF_code_hits table
#
	if( $status ) {
		my $update = $dbh->prepare('UPDATE RF_code_hits SET Num_of_hits = ? WHERE RF_id = ?');
		if( $dbh->err) {
			$lastError = $dbh->errstr;
			$status = 0;
		}
		else {
			my $hits = $stmt->fetchrow_arrayref();
			while($hits && !$update->err) {
				$update->execute(@$hits);
				$hits = $stmt->fetchrow_arrayref();
			}
			if( !$update->err ) {
				$result = 1;
			}
			else {
				$lastError = $update->errstr;
				$status = 0;
			}
		}
		$update->finish();
		$stmt->finish();
	}

	return $result;

}
	
=head1 checkEnvironment - check form required parameters and environment vars

Input

	List of environment variables

Results

	If variable is not present, print a pre-ordained error message
	A false (0) is returned if one or more is missing
=cut

sub checkEnvironment {

	my $result = 1;
	while(@_) {

		my $var = shift;
		if( !exists($ENV{$var}) ) {
			print STDERR $var;
			print STDERR $_ENVERR{$var},"\n";
			$result = 0;
		}

	}

	return $result;

}

=head1 findEntity - lookup ids for given classes and types

Input

	database handle reference
	list of entity names, entity table pairs

	where entity is a class or type and entity table is from what
	table the entity name is expected to be found

Results

	the entity names are looked up to retrieve code

Return value

	list of entity ids
	list will be smaller than expected if some entities are not found

=cut

sub findEntity {

	my $dbh = shift;

	my @result = ();

	while(@_) {
		my $entity = shift;
		my $table = shift;
		my $stmt = $dbh->prepare("SELECT Entity_id FROM $table WHERE UCASE(Entity_name) = ?");
		if( $dbh->err ) {
			$lastError = $dbh->errstr;
		}
		else {
			$stmt->execute(uc $entity);
			if( $stmt->err ) {
				$lastError = $stmt->errstr;
			}
			if( my $classref = $stmt->fetchrow_arrayref() ) {
				$result[@result] = $$classref[0];
			}
			else {
				$lastError = "No Such Name $entity in $table";
			}
		}
		$stmt->finish();
	}

	return @result;

}

=head1 execSQLstmt - execute a single SQL statement

Input

	The database handle used to execute the statement
	The SQL statement as a string
	The remainder is the list used as parameters of SQL statement

Result

	The statement is prepared and executed

Return Value

	undef if not successful execution
	a reference to a list containing references to lists which are
	rows returned from the query
	an empty reference is returned on non-SELECT or empty queries

=cut

sub execSQLstmt {

	my $dbh = shift;
	my $sqlcmd = shift;

	my $result; # undefined until we get results

	my $stmt = $dbh->prepare($sqlcmd);
	if( $dbh->err ) {
		$lastError = $dbh->errstr;
	}
	else {
		$stmt->execute(@_);
		if( $sqlcmd =~ /^SELECT/ ) {
			if( $stmt->err ) {
				$lastError = $stmt->errstr;
			}
			else {
				$result = [];
				while( my $row = $stmt->fetchrow_arrayref()) {
					my @copy = @$row;
					$$result[@$result] = \@copy;
				}
				if( scalar @$result == 0) {
					$result = undef;
				}
			}
		}
	}

	return $result;

}

=head1 getLocation - get location of one feature

Input

	A Bio::Annotation::Generic object

Result

	all intervals of the feature a computed

Returned Value

	a list of array references each with
	two elements representing the intervals of the feature
	the start (first) element of each pair is always
	the closest to the 5' end of the sequence (i.e. if feature
        is on the anti-sense strand


sub getLocation {

	my $annotation = shift;
	my $placement = $annotation->location();
	my @interval; # list to return
	if( $placement->isa('Bio::Location::Split') ) {
		my @subs = $placement->sub_Location();
		foreach my $subfeature (@subs) {
			if( int($annotation->strand()) < 0 ) {
				$interval[@interval] = [$subfeature->end(),$subfeature->start()];
			}
			else {
				$interval[@interval] = [$subfeature->start(),$subfeature->end()];
			}
		} 
	}
	else {
		if( $annotation->strand() < 0 ) {
			$interval[@interval] = [$annotation->end(),$annotation->start()];
		}
		else {
			$interval[@interval] = [$annotation->start(),$annotation->end()];
		}
	}

	return @interval;

}


=head1 getFeatures - get intervals of features in sequence

Input

	Bio::Seq object

Results

	Features are extracted and sub features extracted from features
	A hash map from feature name to a reference to an array
	containing references of a 2-member array each describing
	the start/end of an interval for the corresponding feature

Results

	The hash created is passed back as a reference. The intervals are
	guaranteed to be in order for each feature


sub getFeatures {

my $seqobj = shift;
my @features = $seqobj->top_SeqFeatures();
my %feature_map; # returned feature -> interval array map

	for( my $i=0; $i < @features; $i++) {

		my $feature_name = $features[$i]->primary_tag();
		if( exists($feature_map{$feature_name})) {
			my $named_feature = $feature_map{$feature_name};
			$$named_feature[@$named_feature] = $features[$i];
		}
		else {
			$feature_map{$feature_name} = [$features[$i]];
		}
	}

	foreach my $name (keys %feature_map) {

		my $named_feature = $feature_map{$name};
		my @interval= ();
		for( my $i=0; $i < @$named_feature; $i++) {
			my @places = getLocation($$named_feature[$i]);
			foreach my $place ( @places ) {
				$interval[@interval] = $place;
			}
		}
# sort by position of first base independent of strand
		my @sorted_interval = sort { ($$a[0] < $$a[1] ? $$a[0] : $$a[1]) <=> ($$b[0] < $$b[1] ? $$b[0] : $$b[1]) } @interval;
		$feature_map{$name} = \@sorted_interval; # replace features with intervals
		

	}

	return \%feature_map;

}

=cut
=head1 findPath - find file in path

Input

	filename to find
	an alternate environment variable besides PATH may be given

Results

	The path is searched for the file until found

Returned value

	Full path name of file if found, otherwise filename is returned

=cut

sub findPath {

	my $filename = shift; # get file to lookup
	my $pathvar = 'PATH';

	if( @_ ) { # alternate path string passed
		$pathvar = shift;
	}

	my $directories = $ENV{$pathvar};

	foreach my $dir ( split /:|;| /, $directories ) {
		if( -r $dir.'/'.$filename ) {
			$filename = $dir.'/'.$filename;
			last;
		}
	}

	return $filename;

}

return 1;
